"""KCH Driver."""
from enum import IntEnum, unique
from typing import List, Tuple, Union
from warnings import catch_warnings

try:
    import RPi.GPIO as GPIO  # isort: ignore
    HAS_HAT = True
except ImportError:
    HAS_HAT = False


@unique
class RobotLEDs(IntEnum):
    """Mapping of LEDs to GPIO Pins."""

    START = 9

    USER_A_RED = 24
    USER_A_GREEN = 10
    USER_A_BLUE = 25
    USER_B_RED = 27
    USER_B_GREEN = 23
    USER_B_BLUE = 22
    USER_C_RED = 4
    USER_C_GREEN = 18
    USER_C_BLUE = 17

    @classmethod
    def all_leds(cls) -> List[int]:
        """Get all LEDs."""
        return [c.value for c in cls]

    @classmethod
    def user_leds(cls) -> List[Tuple[int, int, int]]:
        """Get the user programmable LEDs."""
        return [
            (cls.USER_A_RED, cls.USER_A_GREEN, cls.USER_A_BLUE),
            (cls.USER_B_RED, cls.USER_B_GREEN, cls.USER_B_BLUE),
            (cls.USER_C_RED, cls.USER_C_GREEN, cls.USER_C_BLUE),
        ]


@unique
class UserLED(IntEnum):
    """User Programmable LEDs."""

    A = 0
    B = 1
    C = 2


class Colour():
    """User LED colours."""

    OFF = (False, False, False)
    RED = (True, False, False)
    YELLOW = (True, True, False)
    GREEN = (False, True, False)
    CYAN = (False, True, True)
    BLUE = (False, False, True)
    MAGENTA = (True, False, True)
    WHITE = (True, True, True)


class KCH:
    """KCH Board."""

    def __init__(self) -> None:
        if HAS_HAT:
            GPIO.setmode(GPIO.BCM)
            with catch_warnings():
                # If this is not the first time the code is run this init will
                # cause a warning as the gpio are alrady initilized, we can
                # suppress this as we know the reason behind the warning
                GPIO.setup(RobotLEDs.all_leds(), GPIO.OUT, initial=GPIO.LOW)
        self._leds = LEDs(RobotLEDs.user_leds())

    def __del__(self) -> None:
        # We are not running cleanup so the LED state persits after the code completes,
        # this will cause a warning with `GPIO.setup()` which we can suppress
        if HAS_HAT:
            # GPIO.cleanup()
            pass

    @property
    def start(self) -> bool:
        """Get the state of the start LED."""
        return GPIO.input(RobotLEDs.START) if HAS_HAT else False

    @start.setter
    def start(self, value: bool) -> None:
        """Set the state of the start LED."""
        if HAS_HAT:
            GPIO.output(RobotLEDs.START, GPIO.HIGH if value else GPIO.LOW)

    @property
    def leds(self) -> 'LEDs':
        """User programmable LEDs."""
        return self._leds


class LEDs:
    """Programmable LEDs controller."""

    def __init__(self, leds: List[Tuple[int, int, int]]):
        self._leds = leds

    def __setitem__(self, key: Union[int, slice], value: Tuple[bool, bool, bool]) -> None:
        """
        Set the colour of a user LED.

        :param key: The index of the RGB LED to set.
        :param value: A 3-tuple of the subpixel values as bools.
        :returns: None.
        :raises ValueError: value wasn't a tuple with 3 elements.
        :raises IndexError: An index outside range(3) was requested.
        """
        rgb_leds = []
        if isinstance(key, slice):
            rgb_leds = self._leds[key]
        elif isinstance(key, int):
            rgb_leds = [self._leds[key]]

        if not isinstance(value, (tuple, list)) or len(value) != 3:
            raise ValueError("The LED requires 3 values for it's colour")

        if HAS_HAT:
            for rgb_led in rgb_leds:
                for led, state in zip(rgb_led, value):
                    GPIO.output(led, GPIO.HIGH if state else GPIO.LOW)

    def __getitem__(self, key: Union[int, slice]) -> Union['LED', List['LED']]:
        """
        Get the colour of a user LED.

        :param key: The index of the RGB LED to get the colour of.
        :returns: A 3-tuple of the subpixel values as bools.
        :raises IndexError: An index outside range(3) was requested.
        """
        if isinstance(key, slice):
            return [LED(led) for led in self._leds[key]]
        elif isinstance(key, int):
            return LED(self._leds[key])

    def __len__(self) -> int:
        return len(self._leds)


class LED:
    """User programmable LED."""

    def __init__(self, led: Tuple[int, int, int]):
        self._led = led

    @property
    def r(self) -> bool:
        """Get the state of the Red LED segment."""
        return GPIO.input(self._led[0]) if HAS_HAT else False

    @r.setter
    def r(self, value: bool) -> None:
        """Set the state of the Red LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[0], GPIO.HIGH if value else GPIO.LOW)

    @property
    def g(self) -> bool:
        """Get the state of the Green LED segment."""
        return GPIO.input(self._led[1]) if HAS_HAT else False

    @g.setter
    def g(self, value: bool) -> None:
        """Set the state of the Green LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[1], GPIO.HIGH if value else GPIO.LOW)

    @property
    def b(self) -> bool:
        """Get the state of the Blue LED segment."""
        return GPIO.input(self._led[2]) if HAS_HAT else False

    @b.setter
    def b(self, value: bool) -> None:
        """Set the state of the Blue LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[2], GPIO.HIGH if value else GPIO.LOW)



"""Game specific code."""
from typing import Container, Dict


class UnusedMarkerException(Exception):
    """This marker is not used for the game."""


MARKER_SIZES: Dict[Container[int], int] = {
    range(28): 200,  # 0 - 27 for arena boundary
    range(28, 100): 100,  # Everything else is a token
}


def marker_used_in_game(marker_id: int) -> bool:
    """
    Determine whether the marker ID is used in the game.

    :param marker_id: An official marker number, mapped to the competitor range.
    :returns: True if the market is used in the game.
    """
    return any([marker_id in marker_range for marker_range in MARKER_SIZES])


def get_marker_size(marker_id: int) -> int:
    """
    Get the size of a marker in millimetres.

    :param marker_id: An official marker number, mapped to the competitor range.
    :returns: Size of the marker in millimetres.
    """
    for marker_range, size in MARKER_SIZES.items():
        if marker_id in marker_range:
            return size
    raise UnusedMarkerException(f"{marker_id} is not used for the game.")



"""Strategies for determining the camera on given hardware."""
import json
import platform
import subprocess
from abc import ABCMeta, abstractmethod
from typing import Optional

import usb


class CalibrationStrategy(metaclass=ABCMeta):
    """Base calibration strategy used to determine the correct camera calibration file."""

    @abstractmethod
    def get_calibration_name(self) -> Optional[str]:
        """Get the name of the calibration profile."""
        raise NotImplementedError


class StaticStrategy(CalibrationStrategy):
    """A strategy that always returns the same calibration profile."""

    def __init__(self, calibration_name: str) -> None:
        self._calibration_name = calibration_name

    def get_calibration_name(self) -> Optional[str]:
        """Get the name of the calibration profile."""
        return self._calibration_name


class USBDevicePresentStrategy(CalibrationStrategy):
    """Strategy used for determining calibration profile based on present USB devices."""

    def __init__(
            self,
            vendor_id: int,
            product_id: int,
            calibration_name: str,
    ) -> None:
        self._vendor_id = vendor_id
        self._product_id = product_id
        self._calibration_name = calibration_name

    def get_calibration_name(self) -> Optional[str]:
        """Gets the name of the calibration profile."""
        if usb.core.find(
                idVendor=self._vendor_id,
                idProduct=self._product_id,
        ) is not None:
            return self._calibration_name
        else:
            return None


class MacSystemStrategy(CalibrationStrategy):
    """Strategy used for determining calibration profile on macOS."""

    def get_calibration_name(self) -> Optional[str]:
        """Get the name of the calibration profile."""
        if platform.system() == "Darwin":  # macOS
            return self._find_default_camera()
        return None

    def _find_default_camera(self) -> Optional[str]:
        camera_info = json.loads(
            subprocess.check_output(['system_profiler', '-json', 'SPCameraDataType']),
        )
        first_camera = camera_info['SPCameraDataType'][0]
        if first_camera['_name'] == 'FaceTime HD Camera':
            return 'FaceTime HD Camera'
        return None



"""
SR custom behaviour for Zoloto.

- Overriding the camera so that only the markers we want are captured
- Map the development and competition markers into the correct ranges
- Determine the size of a marker given the ID
- Determine which OpenCV calibration to use for the currently connected camera.
"""
import logging
from pathlib import Path
from typing import List, Optional, Tuple

from j5_zoloto import ZolotoHardwareBackend
from numpy.typing import NDArray
from zoloto.calibration import parse_calibration_file
from zoloto.cameras import Camera
from zoloto.marker_type import MarkerType

from sr.robot3.game import get_marker_size, marker_used_in_game

from .strategy import (
    CalibrationStrategy,
    MacSystemStrategy,
    USBDevicePresentStrategy,
)

LOGGER = logging.getLogger(__name__)

STRATEGIES: List[CalibrationStrategy] = [
    USBDevicePresentStrategy(0x046d, 0x0825, "Logitech C270"),
    USBDevicePresentStrategy(0x046d, 0x0807, "Logitech B500"),
    USBDevicePresentStrategy(0x046d, 0x080a, "Logitech C905"),
    USBDevicePresentStrategy(0x046d, 0x082d, "Logitech C920"),
    USBDevicePresentStrategy(0x046d, 0x0892, "Logitech C920"),
    USBDevicePresentStrategy(0x046d, 0x08E5, "Logitech C920"),  # C920 PRO
    USBDevicePresentStrategy(0x0c45, 0x6713, "Microdia Integrated_Webcam_HD"),
    MacSystemStrategy(),
]


class SRZolotoCamera(Camera):
    """A Zoloto camera that correctly captures markers for SR."""

    def __init__(
        self,
        camera_id: int,
        *,
        marker_size: Optional[int] = None,
        marker_type: MarkerType,
        calibration_file: Optional[Path] = None,
        marker_offset: int = 0,
    ) -> None:
        resolution: Optional[Tuple[int, int]] = None

        if calibration_file is not None:
            resolution = parse_calibration_file(calibration_file).resolution

        super().__init__(
            camera_id,
            marker_size=marker_size,
            marker_type=marker_type,
            calibration_file=calibration_file,
            resolution=resolution,
        )
        self._marker_offset = marker_offset

    def _get_ids_and_corners(
        self, frame: Optional[NDArray] = None,  # type: ignore[type-arg]
    ) -> Tuple[List[int], List[NDArray]]:  # type: ignore[type-arg]
        raw_ids, raw_corners = super()._get_ids_and_corners(frame)

        ids: List[int] = []
        corners: List[NDArray] = []   # type: ignore[type-arg]

        # Copy list, map marker IDs and filter out ones not in game
        for raw_id, raw_corner in zip(raw_ids, raw_corners):
            offset_id = raw_id - self._marker_offset
            if marker_used_in_game(offset_id):
                ids.append(offset_id)
                corners.append(raw_corner)

        return ids, corners

    def get_marker_size(self, marker_id: int) -> int:
        """
        Get the size of a marker given it's ID.

        :param marker_id: The offical ID number of the marker.
        :returns: The size of the marker in millimetres.
        """
        return get_marker_size(marker_id)


class SRZolotoHardwareBackend(ZolotoHardwareBackend):
    """A camera backend which automatically finds camera calibration data."""

    camera_class = SRZolotoCamera
    marker_type = MarkerType.APRILTAG_36H11

    def __init__(self, camera_id: int, *, marker_offset: int = 0) -> None:
        self._zcam = self.camera_class(
            camera_id,
            marker_type=self.marker_type,
            calibration_file=self.get_calibration_file(),
            marker_size=self.marker_size,
            marker_offset=marker_offset,
        )

    def get_calibration_file(self) -> Optional[Path]:
        """Get the calibration file to use."""
        for strategy in STRATEGIES:
            filename = strategy.get_calibration_name()
            if filename is not None:
                LOGGER.debug(f"Using {filename} for webcam calibration")
                return Path(__file__).parent / f'calibrations/{filename}.xml'

        LOGGER.warning(
            "Unable to determine camera calibration, disabling pose estimation",
        )
        return None
